# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016 CERN.
# Copyright (C) 2020 data-futures.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Multimedia Image API Tests."""

from io import BytesIO

import pytest
from PIL import Image

from flask_iiif.api import MultimediaImage

from .helpers import IIIFTestCase


class TestMultimediaAPI(IIIFTestCase):
    """Multimedia Image API test case."""

    @pytest.mark.parametrize(
        "width, height",
        [(1280, 1024), (100, 100), (1024, 1280), (200, 100), (1280, 720)],  # portrait
    )
    def setUp(self, width=1280, height=1024):
        """Run before the test."""
        # Create an image in memory

        tmp_file = BytesIO()
        # create a new image
        image = Image.new("RGBA", (width, height), (255, 0, 0, 0))
        image.save(tmp_file, "png")
        self.width, self.height = width, height

        # Initialize it for our object and create and instance for
        # each test
        tmp_file.seek(0)
        self.image_resize = MultimediaImage.from_string(tmp_file)
        tmp_file.seek(0)
        self.image_crop = MultimediaImage.from_string(tmp_file)
        tmp_file.seek(0)
        self.image_rotate = MultimediaImage.from_string(tmp_file)
        tmp_file.seek(0)
        self.image_errors = MultimediaImage.from_string(tmp_file)
        tmp_file.seek(0)
        self.image_convert = MultimediaImage.from_string(tmp_file)
        tmp_file.seek(0)
        self.image_save = MultimediaImage.from_string(tmp_file)

        # NOT RGBA
        tmp_file = BytesIO()
        # create a new image
        image = Image.new("RGB", (width, height), (255, 0, 0, 0))
        image.save(tmp_file, "jpeg")
        tmp_file.seek(0)
        self.image_not_rgba = MultimediaImage.from_string(tmp_file)

        # Image in P Mode
        tmp_file = BytesIO()
        image = Image.new("P", (width, height))
        image.save(tmp_file, "gif")
        tmp_file.seek(0)
        self.image_p_mode = MultimediaImage.from_string(tmp_file)

        # TIFF Image
        tmp_file = BytesIO()
        image = Image.new("RGBA", (width, height), (255, 0, 0, 0))
        image.save(tmp_file, "tiff")
        tmp_file.seek(0)
        self.image_tiff = MultimediaImage.from_string(tmp_file)

    def test_image_resize(self):
        """Test image resize function."""
        # Test image size before
        self.assertEqual(str(self.image_resize.size()), str((self.width, self.height)))

        # Resize.image_resize to 720,680
        self.image_resize.resize("720,680")
        self.assertEqual(str(self.image_resize.size()), str((720, 680)))

        # Resize.image_resize to 300,
        self.image_resize.resize("300,")
        self.assertEqual(str(self.image_resize.size()), str((300, 283)))

        # Resize.image_resize to ,300
        self.image_resize.resize(",300")
        self.assertEqual(str(self.image_resize.size()), str((318, 300)))

        # Resize.image_resize to pct:90
        self.image_resize.resize("pct:90")
        self.assertEqual(str(self.image_resize.size()), str((286, 270)))

        # Resize.image_resize to !100,100
        self.image_resize.resize("!100,100")
        self.assertEqual(str(self.image_resize.size()), str((100, 94)))

    def test_errors(self):
        """Test errors."""
        from flask_iiif.errors import (
            MultimediaImageCropError,
            MultimediaImageFormatError,
            MultimediaImageNotFound,
            MultimediaImageQualityError,
            MultimediaImageResizeError,
        )

        # Test resize errors
        self.assertRaises(
            MultimediaImageResizeError, self.image_errors.resize, "pct:-12222"
        )
        self.assertRaises(MultimediaImageResizeError, self.image_errors.resize, "2")
        self.assertRaises(
            MultimediaImageResizeError, self.image_errors.resize, "-22,100"
        )
        # Test crop errors
        self.assertRaises(
            MultimediaImageCropError, self.image_errors.crop, "22,100,222"
        )
        self.assertRaises(
            MultimediaImageCropError, self.image_errors.crop, "-22,100,222,323"
        )
        self.assertRaises(
            MultimediaImageCropError, self.image_errors.crop, "pct:222,100,222,323"
        )
        self.assertRaises(
            MultimediaImageCropError, self.image_errors.crop, "2000,2000,2000,2000"
        )
        self.assertRaises(
            MultimediaImageNotFound, self.image_errors.from_file, "unicorn"
        )
        with self.app.app_context():
            self.assertRaises(
                MultimediaImageQualityError,
                self.image_errors.quality,
                "pct:222,100,222,323",
            )
            self.assertRaises(
                MultimediaImageFormatError,
                self.image_errors._prepare_for_output,
                "pct:222,100,222,323",
            )

    def test_image_crop(self):
        """Test the crop function."""
        # Crop image x,y,w,h
        self.image_crop.crop("20,20,400,300")
        self.assertEqual(str(self.image_crop.size()), str((400, 300)))

        # Crop image pct:x,y,w,h
        self.image_crop.crop("pct:20,20,40,30")
        self.assertEqual(str(self.image_crop.size()), str((160, 90)))

        # Check if exeeds image borders
        self.image_crop.crop("10,10,160,90")
        self.assertEqual(str(self.image_crop.size()), str((150, 80)))

    def test_image_rotate(self):
        """Test image rotate function."""
        self.image_rotate.rotate(90)
        self.assertEqual(str(self.image_rotate.size()), str((1024, 1280)))

        self.image_rotate.rotate(120)
        self.assertEqual(str(self.image_rotate.size()), str((1622, 1528)))

    def test_image_mode(self):
        """Test image mode."""
        self.image_not_rgba.quality("grey")
        self.assertEqual(self.image_not_rgba.image.mode, "L")

    def test_image_incompatible_modes(self):
        """Test P-incompatible image to RGB auto-convert."""
        tmp_file = BytesIO()
        self.image_p_mode.save(tmp_file)
        self.assertEqual(self.image_p_mode.image.mode, "RGB")

    def test_image_saving(self):
        """Test image saving."""
        tmp_file = BytesIO()
        compare_file = BytesIO()
        self.assertEqual(tmp_file.getvalue(), compare_file.getvalue())
        self.image_save.save(tmp_file)
        self.assertNotEqual(str(tmp_file.getvalue()), compare_file.getvalue())

    def test_image_tiff_support(self):
        """Test TIFF image support."""
        self.assertEqual(self.image_tiff.image.format, "TIFF")



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.
r"""Flask-IIIF tests."""



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016 CERN.
# Copyright (C) 2020 data-futures.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Flask-IIIF utilities."""
import datetime
import shutil
import tempfile
from email.utils import formatdate
from os.path import dirname, join

from flask import abort, current_app, url_for
from PIL import Image, ImageSequence

__all__ = ("iiif_image_url",)


def iiif_image_url(**kwargs):
    """Generate a `IIIF API` image url.

    :returns: `IIIF API` image url
    :rtype: str

    .. code:: html

        <img
            src="{{ iiif_image_url(uuid="id", size="200,200") }}"
            alt="Title"
        />

    .. note::

        If any IIIF parameter missing it will fall back to default:
        * `image_format = png`
        * `quality = default`
        * `region = full`
        * `rotation = 0`
        * `size = full`
        * `version = v2`
    """
    try:
        assert kwargs.get("uuid") is not None
    except AssertionError:
        abort(404)
    else:
        url_for_args = {k: v for k, v in kwargs.items() if k.startswith("_")}

        return url_for(
            "iiifimageapi",
            image_format=kwargs.get("image_format", "png"),
            quality=kwargs.get("quality", "default"),
            region=kwargs.get("region", "full"),
            rotation=kwargs.get("rotation", 0),
            size=kwargs.get("size", "full"),
            uuid=kwargs.get("uuid"),
            version=kwargs.get("version", "v2"),
            **url_for_args
        )


def should_cache(request_args):
    """Check the request args for cache-control specifications.

    :param request_args: flask request args
    """
    if "cache-control" in request_args and request_args["cache-control"] in [
        "no-cache",
        "no-store",
    ]:
        return False
    return True


def datetime_to_float(date):
    """Convert datetime to string accepted by browsers as per RFC 2822."""
    epoch = datetime.datetime.utcfromtimestamp(0)
    total_seconds = (date - epoch).total_seconds()
    return formatdate(total_seconds, usegmt=True)



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2015 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Test REST API."""

from io import BytesIO
from unittest.mock import patch

from flask import url_for
from PIL import Image
from werkzeug.utils import secure_filename

from .helpers import IIIFTestCase


class TestRestAPI(IIIFTestCase):

    """Test signals and decorators."""

    def test_api_base(self):
        """Test API Base."""
        data = dict(uuid="valid:id", version="v2")
        get_the_response = self.get("iiifimagebase", urlargs=data)
        self.assertEqual(get_the_response.status_code, 303)

    def test_api_info(self):
        """Test API Info not found case."""
        from flask import jsonify

        id_v1 = url_for(
            "iiifimagebase", uuid="valid:id-üni", version="v1", _external=True
        )
        id_v2 = url_for(
            "iiifimagebase", uuid="valid:id-üni", version="v2", _external=True
        )

        expected = {
            "v1": {
                "@context": (
                    "http://library.stanford.edu/iiif/" "image-api/1.1/context.json"
                ),
                "@id": id_v1,
                "width": 1280,
                "height": 1024,
                "profile": (
                    "http://library.stanford.edu/iiif/image-api/compliance"
                    ".html#level1"
                ),
                "tile_width": 256,
                "tile_height": 256,
                "scale_factors": [1, 2, 4, 8, 16, 32, 64],
            },
            "v2": {
                "@context": "http://iiif.io/api/image/2/context.json",
                "@id": id_v2,
                "protocol": "http://iiif.io/api/image",
                "width": 1280,
                "height": 1024,
                "tiles": [{"width": 256, "scaleFactors": [1, 2, 4, 8, 16, 32, 64]}],
                "profile": ["http://iiif.io/api/image/2/level2.json"],
            },
        }
        get_the_response = self.get(
            "iiifimageinfo",
            urlargs=dict(
                uuid="valid:id-üni",
                version="v2",
            ),
        )
        self.assert200(get_the_response)
        get_the_response = self.get(
            "iiifimageinfo",
            urlargs=dict(
                uuid="valid:id-üni",
                version="v2",
            ),
        )
        self.assertEqual(jsonify(expected.get("v2")).data, get_the_response.data)

        get_the_response = self.get(
            "iiifimageinfo",
            urlargs=dict(
                uuid="valid:id-üni",
                version="v1",
            ),
        )
        self.assert200(get_the_response)
        self.assertEqual(jsonify(expected.get("v1")).data, get_the_response.data)

    def test_api_info_not_found(self):
        """Test API Info."""
        get_the_response = self.get(
            "iiifimageinfo",
            urlargs=dict(
                uuid="notfound",
                version="v2",
            ),
        )
        self.assert404(get_the_response)

    def test_api_not_found(self):
        """Test API not found case."""
        get_the_response = self.get(
            "iiifimageapi",
            urlargs=dict(
                uuid="notfound",
                version="v2",
                region="full",
                size="full",
                rotation="0",
                quality="default",
                image_format="png",
            ),
        )
        self.assert404(get_the_response)

    def test_api_internal_server_error(self):
        """Test API internal server error case."""
        get_the_response = self.get(
            "iiifimageapi",
            urlargs=dict(
                uuid="valid:id-üni",
                version="v2",
                region="full",
                size="full",
                rotation="2220",
                quality="default",
                image_format="png",
            ),
        )
        self.assert500(get_the_response)

    def test_api_iiif_validation_error(self):
        """Test API iiif validation case."""
        get_the_response = self.get(
            "iiifimageapi",
            urlargs=dict(
                uuid="valid:id-üni",
                version="v1",
                region="200",
                size="full",
                rotation="2220",
                quality="default",
                image_format="png",
            ),
        )
        self.assert400(get_the_response)

    def test_api_stream_image(self):
        """Test API stream image."""
        tmp_file = BytesIO()
        # create a new image
        image = Image.new("RGBA", (1280, 1024), (255, 0, 0, 0))
        image.save(tmp_file, "png")
        tmp_file.seek(0)
        get_the_response = self.get(
            "iiifimageapi",
            urlargs=dict(
                uuid="valid:id-üni",
                version="v2",
                region="full",
                size="full",
                rotation="0",
                quality="default",
                image_format="png",
            ),
        )
        # Check if returns `Last-Modified` key in headers
        # required for `If-Modified-Since`
        self.assertTrue("Last-Modified" in get_the_response.headers)

        last_modified = get_the_response.headers["Last-Modified"]

        self.assertEqual(get_the_response.data, tmp_file.getvalue())

        # Test `If-Modified-Since` recognized properly
        get_the_response = self.get(
            "iiifimageapi",
            urlargs=dict(
                uuid="valid:id-üni",
                version="v2",
                region="full",
                size="full",
                rotation="0",
                quality="default",
                image_format="png",
            ),
            headers={"If-Modified-Since": last_modified},
        )

        self.assertEqual(get_the_response.status_code, 304)

        urlargs = dict(
            uuid="valid:id-üni",
            version="v2",
            region="200,200,200,200",
            size="300,300",
            rotation="!50",
            quality="color",
            image_format="pdf",
        )

        get_the_response = self.get(
            "iiifimageapi",
            urlargs=urlargs,
        )
        self.assert200(get_the_response)

        default_name = "{name}-200200200200-300300-color-50.pdf".format(
            name=secure_filename(urlargs["uuid"])
        )
        for dl, name in (
            ("", default_name),
            ("1", default_name),
            ("foo.pdf", "foo.pdf"),
        ):
            urlargs["dl"] = dl
            get_the_response = self.get(
                "iiifimageapi",
                urlargs=urlargs,
            )
            self.assert200(get_the_response)
            self.assertEqual(
                get_the_response.headers["Content-Disposition"],
                "attachment; filename={name}".format(name=name),
            )

    def test_api_decorator(self):
        """Test API decorator."""
        get_the_response = self.get(
            "iiifimageapi",
            urlargs=dict(
                uuid="valid:decorator:id",
                version="v2",
                region="full",
                size="full",
                rotation="0",
                quality="default",
                image_format="png",
            ),
        )
        self.assert403(get_the_response)

    def test_api_abort_all_methods_except_get(self):
        """Abort all methods but GET."""
        data = dict(
            uuid="valid:id-üni",
            version="v2",
            region="full",
            size="full",
            rotation="0",
            quality="default",
            image_format="png",
        )
        get_the_response = self.post("iiifimageapi", urlargs=data)
        self.assert405(get_the_response)

        get_the_response = self.put("iiifimageapi", urlargs=data)
        self.assert405(get_the_response)

        get_the_response = self.delete("iiifimageapi", urlargs=data)
        self.assert405(get_the_response)

        get_the_response = self.patch("iiifimageapi", urlargs=data)
        self.assert405(get_the_response)

    def test_api_cache_control(self):
        """Test cache-control headers"""

        urlargs = dict(
            uuid="valid:id-üni",
            version="v2",
            region="200,200,200,200",
            size="300,300",
            rotation="!50",
            quality="color",
            image_format="pdf",
        )

        key = "iiif:{0}/{1}/{2}/{3}/{4}.{5}".format(
            urlargs["uuid"],
            urlargs["region"],
            urlargs["size"],
            urlargs["quality"],
            urlargs["rotation"],
            urlargs["image_format"],
        )

        cache = self.app.config["IIIF_CACHE_HANDLER"].cache

        get_the_response = self.get(
            "iiifimageapi",
            urlargs=urlargs,
        )

        self.assertFalse("cache-control" in urlargs)

        self.assert200(get_the_response)

        self.assertTrue(cache.get(key))

        cache.clear()

        for cache_control, name in (("no-cache", "foo.pdf"), ("no-store", "foo.pdf")):
            urlargs["cache-control"] = cache_control

            get_the_response = self.get(
                "iiifimageapi",
                urlargs=urlargs,
            )

            self.assert200(get_the_response)

            self.assertFalse(cache.get(key))

            cache.clear()

        for cache_control, name in (("public", "foo.pdf"), ("no-transform", "foo.pdf")):
            urlargs["cache-control"] = cache_control

            get_the_response = self.get(
                "iiifimageapi",
                urlargs=urlargs,
            )

            self.assert200(get_the_response)

            self.assertTrue(cache.get(key))

            cache.clear()

    def test_cache_ignore_errors(self):
        """Test if cache retrieval errors are ignored when configured."""
        from flask import current_app

        info_args = dict(uuid="valid:id", version="v2")
        api_args = dict(
            uuid="valid:id",
            version="v2",
            region="full",
            size="full",
            rotation="0",
            quality="default",
            image_format="png",
        )

        cache = self.app.config["IIIF_CACHE_HANDLER"].cache
        with patch.object(cache, "get", side_effect=Exception("test fail")):
            # Without ignoring errors
            self.assertRaisesRegex(
                Exception, "test fail", self.get, "iiifimageinfo", urlargs=info_args
            )
            self.assertRaisesRegex(
                Exception, "test fail", self.get, "iiifimageapi", urlargs=api_args
            )

            # Ignore errors
            old_value = current_app.config.get("IIIF_CACHE_IGNORE_ERRORS")
            current_app.config["IIIF_CACHE_IGNORE_ERRORS"] = True

            resp = self.get("iiifimageinfo", urlargs=info_args)
            self.assert200(resp)
            resp = self.get("iiifimageapi", urlargs=api_args)
            self.assert200(resp)

            current_app.config["IIIF_CACHE_REDIS_PREFIX"] = old_value



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016, 2017 CERN.
# Copyright (C) 2020 data-futures.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""IIIF configuration.

.. py:data:: IIIF_CACHE_HANDLER

    Add the preferred cache adaptor.

    .. seealso:: :py:class:`~flask_iiif.cache.cache.ImageCache`

.. py:data:: IIIF_CACHE_REDIS_PREFIX

    Sets prefix for redis keys, default: `iiif`

.. py:data:: IIIF_CACHE_TIME

    How much time the image would be cached.

.. py:data:: IIIF_QUALITIES

    The supported image qualities.

    .. seealso::

            `IIIF Image API
            <http://iiif.io/api/image/2.0/#quality>`_

.. py:data:: IIIF_CONVERTERS

    The supported image converters.

.. py:data:: IIIF_FORMATS

    The supported image formats with their MIME type.

.. py:data:: IIIF_VALIDATIONS

    The IIIF Image API validation.

    .. seealso::

        `IIIF Image API v1
        <http://iiif.io/api/image/1.1/>`_ and
        `IIIF Image API v2
        <http://iiif.io/api/image/2.0/>`_

.. py:data:: IIIF_API_INFO_RESPONSE_SKELETON

    Information request document for the image.

    .. seealso::
        `IIIF Image API v1 Information request
        <http://iiif.io/api/image/1.1/#information-request>`_ and
        `IIIF Image API v2 Information request
        <http://iiif.io/api/image/2.0/#information-request>`_

"""
# Cache handler
IIIF_CACHE_HANDLER = "flask_iiif.cache.simple:ImageSimpleCache"

# Cache duration
# 60 seconds * 60 minutes (1 hour) * 24 (24 hours) * 2 (2 days) = 172800 secs
# 60 seconds * 60 (1 hour) * 24 (1 day) * 2 (2 days)
IIIF_CACHE_TIME = 60 * 60 * 24 * 2

# Redis URL Cache
IIIF_CACHE_REDIS_URL = "redis://localhost:6379/0"

# Supported qualities
IIIF_QUALITIES = ("default", "gray", "grey", "bitonal", "color", "native")
# Suported coverters
IIIF_CONVERTERS = "", "L", "L", "1", "", ""

# Supported multimedia image API formats
IIIF_FORMATS = {
    "gif": "image/gif",
    "jp2": "image/jp2",
    "jpeg": "image/jpeg",
    "jpg": "image/jpeg",
    "pdf": "application/pdf",
    "png": "image/png",
    "tif": "image/tiff",
    "tiff": "image/tiff",
}

IIIF_FORMATS_PIL_MAP = {
    "gif": "gif",
    "jp2": "jpeg2000",
    "jpeg": "jpeg",
    "jpg": "jpeg",
    "pdf": "pdf",
    "png": "png",
    "tif": "tiff",
    "tiff": "tiff",
}

# Regular expressions to validate each parameter
IIIF_VALIDATIONS = {
    "v1": {
        "region": {
            "ignore": "full",
            "validate": "(^full|(pct:)?([\d.]+,){3}([\d.]+))",
        },
        "size": {
            "ignore": "full",
            "validate": (
                "(^full|[\d.]+,|,[\d.]+|pct:[\d.]+|[\d.]+," "[\d.]+|![\d.]+,[\d.]+)"
            ),
        },
        "rotation": {"ignore": "0", "validate": "^[\d.]+$"},
        "quality": {"ignore": "default", "validate": "(native|color|gr[ae]y|bitonal)"},
        "image_format": {"ignore": "", "validate": "(gif|jp2|jpe?g|pdf|png|tiff?)"},
    },
    "v2": {
        "region": {
            "ignore": "full",
            "validate": "(^full|(pct:)?([\d.]+,){3}([\d.]+))",
        },
        "size": {
            "ignore": "full",
            "validate": (
                "(^full|[\d.]+,|,[\d.]+|pct:[\d.]+|[\d.]+," "[\d.]+|![\d.]+,[\d.]+)"
            ),
        },
        "rotation": {"ignore": "0", "validate": "^!?[\d.]+$"},
        "quality": {"ignore": "default", "validate": "(default|color|gr[ae]y|bitonal)"},
        "image_format": {"ignore": "", "validate": "(gif|jp2|jpe?g|pdf|png|tiff?)"},
    },
}

# Qualities per image mode
IIIF_MODE = {
    "1": ["default", "bitonal"],
    "L": ["default", "gray", "grey", "bitonal"],
    "P": ["default", "gray", "grey", "bitonal"],
    "RGB": ["default", "color", "gray", "grey", "bitonal"],
    "RGBA": ["default", "color", "gray", "grey", "bitonal"],
    "CMYK": ["default", "color", "gray", "grey", "bitonal"],
    "YCbCr": ["default", "color", "gray", "grey", "bitonal"],
    "I": ["default", "color", "gray", "grey", "bitonal"],
    "F": ["default", "color", "gray", "grey", "bitonal"],
}

# API Info
IIIF_API_INFO_RESPONSE_SKELETON = {
    "v1": {
        "@context": ("http://library.stanford.edu/iiif/image-api/1.1/context.json"),
        "@id": "",
        "width": "",
        "height": "",
        "profile": (
            "http://library.stanford.edu/iiif/image-api/compliance.html#level1"
        ),
        "tile_width": 256,
        "tile_height": 256,
        "scale_factors": [1, 2, 4, 8, 16, 32, 64],
    },
    "v2": {
        "@context": "http://iiif.io/api/image/2/context.json",
        "@id": "",
        "protocol": "http://iiif.io/api/image",
        "width": "",
        "height": "",
        "tiles": [{"width": 256, "scaleFactors": [1, 2, 4, 8, 16, 32, 64]}],
        "profile": ["http://iiif.io/api/image/2/level2.json"],
    },
}

# Raise errors during interactions with the cache.
IIIF_CACHE_IGNORE_ERRORS = False

IIIF_GIF_TEMP_FOLDER_PATH = "/tmp"



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016, 2017 CERN.
# Copyright (C) 2020 data-futures.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Multimedia IIIF Image API."""
import datetime
from email.utils import parsedate
from io import BytesIO

from flask import Response, current_app, jsonify, redirect, request, send_file, url_for
from flask_restful import Resource
from flask_restful.utils import cors
from werkzeug.local import LocalProxy
from werkzeug.utils import secure_filename

from .api import IIIFImageAPIWrapper
from .decorators import api_decorator, error_handler
from .signals import (
    iiif_after_info_request,
    iiif_after_process_request,
    iiif_before_info_request,
    iiif_before_process_request,
)
from .utils import should_cache

current_iiif = LocalProxy(lambda: current_app.extensions["iiif"])


class IIIFImageBase(Resource):
    """IIIF Image Base."""

    def get(self, version, uuid):
        """Get IIIF Image Base.

        .. note::

            It will redirect to ``iiifimageinfo`` endpoint with status code
            303.
        """
        return redirect(url_for("iiifimageinfo", version=version, uuid=uuid), code=303)


class IIIFImageInfo(Resource):
    """IIIF Image Info."""

    method_decorators = [
        error_handler,
        api_decorator,
    ]

    @cors.crossdomain(origin="*", methods="GET")
    def get(self, version, uuid):
        """Get IIIF Image Info."""
        # Trigger event before proccess the api request
        iiif_before_info_request.send(self, version=version, uuid=uuid)

        # build the image key
        key = "iiif:info:{0}/{1}".format(version, uuid)

        # Check if its cached
        try:
            cached = current_iiif.cache.get(key)
        except Exception:
            if current_app.config.get("IIIF_CACHE_IGNORE_ERRORS", False):
                cached = None
            else:
                raise

        # If the image size is cached loaded from cache
        if cached:
            width, height = map(int, cached.split(","))
        else:
            data = current_iiif.uuid_to_image_opener(uuid)
            image = IIIFImageAPIWrapper.open_image(data)
            width, height = image.size()
            image.close_image()
            if should_cache(request.args):
                try:
                    current_iiif.cache.set(key, "{0},{1}".format(width, height))
                except Exception:
                    if not current_app.config.get("IIIF_CACHE_IGNORE_ERRORS", False):
                        raise

        data = current_app.config["IIIF_API_INFO_RESPONSE_SKELETON"][version]

        base_uri = url_for("iiifimagebase", uuid=uuid, version=version, _external=True)
        data["@id"] = base_uri
        data["width"] = width
        data["height"] = height

        # Trigger event after proccess the api request
        iiif_after_info_request.send(self, **data)

        resp = jsonify(data)
        if "application/ld+json" in request.headers.get("Accept", ""):
            resp.mimetype = "application/ld+json"
        return resp


class IIIFImageAPI(Resource):
    """IIIF API Implementation.

    .. note::

        * IIF IMAGE API v1.0
            * For more infos please visit <http://iiif.io/api/image/>.
        * IIIF Image API v2.0
            * For more infos please visit <http://iiif.io/api/image/2.0/>.
        * The API works only for GET requests
        * The image process must follow strictly the following workflow:

           * Region
           * Size
           * Rotation
           * Quality
           * Format

    """

    method_decorators = [
        error_handler,
        api_decorator,
    ]

    def get(self, version, uuid, region, size, rotation, quality, image_format):
        """Run IIIF Image API workflow."""
        api_parameters = dict(
            version=version,
            uuid=uuid,
            region=region,
            size=size,
            rotation=rotation,
            quality=quality,
            image_format=image_format,
        )
        # Trigger event before proccess the api request
        iiif_before_process_request.send(self, **api_parameters)

        # Validate IIIF parameters
        IIIFImageAPIWrapper.validate_api(**api_parameters)

        # build the image key
        key = "iiif:{0}/{1}/{2}/{3}/{4}.{5}".format(
            uuid, region, size, quality, rotation, image_format
        )

        # Check if its cached
        try:
            cached = current_iiif.cache.get(key)
        except Exception:
            if current_app.config.get("IIIF_CACHE_IGNORE_ERRORS", False):
                cached = None
            else:
                raise

        # If the image is cached loaded from cache
        if cached:
            to_serve = BytesIO(cached)
            to_serve.seek(0)
        # Otherwise create the image
        else:
            data = current_iiif.uuid_to_image_opener(uuid)
            image = IIIFImageAPIWrapper.open_image(data)

            image.apply_api(
                version=version,
                region=region,
                size=size,
                rotation=rotation,
                quality=quality,
            )

            # prepare image to be serve
            to_serve = image.serve(image_format=image_format)
            image.close_image()
            if should_cache(request.args):
                try:
                    current_iiif.cache.set(key, to_serve.getvalue())
                except Exception:
                    if not current_app.config.get("IIIF_CACHE_IGNORE_ERRORS", False):
                        raise

        try:
            last_modified = current_iiif.cache.get_last_modification(key)
        except Exception:
            if not current_app.config.get("IIIF_CACHE_IGNORE_ERRORS", False):
                raise
            last_modified = None

        # decide the mime_type from the requested image_format
        mimetype = current_app.config["IIIF_FORMATS"].get(image_format, "image/jpeg")
        # Built the after request parameters
        api_after_request_parameters = dict(mimetype=mimetype, image=to_serve)

        # Trigger event after proccess the api request
        iiif_after_process_request.send(self, **api_after_request_parameters)
        send_file_kwargs = {"mimetype": mimetype}
        # last_modified is not supported before flask 0.12
        additional_headers = []
        if last_modified:
            send_file_kwargs.update(last_modified=last_modified)

        if "dl" in request.args:
            filename = secure_filename(request.args.get("dl", ""))
            if filename.lower() in {"", "1", "true"}:
                filename = "{0}-{1}-{2}-{3}-{4}.{5}".format(
                    uuid, region, size, quality, rotation, image_format
                )
            send_file_kwargs.update(
                as_attachment=True,
                download_name=secure_filename(filename),
            )
        if_modified_since_raw = request.headers.get("If-Modified-Since")
        if if_modified_since_raw:
            if_modified_since = datetime.datetime(*parsedate(if_modified_since_raw)[:6])
            if if_modified_since and if_modified_since >= last_modified:
                return Response(status=304)
        response = send_file(to_serve, **send_file_kwargs)
        if additional_headers:
            response.headers.extend(additional_headers)
        return response



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014-2020 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Abstract simple cache definition.

All cache adaptors must at least implement
:func:`~flask_iiif.cache.cache.ImageCache.get` and
:func:`~flask_iiif.cache.cache.ImageCache.set` methods.
"""

from flask import current_app
from werkzeug.utils import cached_property


class ImageCache(object):
    """Abstract cache layer."""

    def __init__(self, app=None):
        """Initialize the cache."""

    @cached_property
    def timeout(self):
        """Return default timeout from config."""
        return current_app.config["IIIF_CACHE_TIME"]

    def get(self, key):
        """Return the key value.

        :param key: the object's key
        """

    def set(self, key, value, timeout=None):
        """Cache the object.

        :param key: the object's key
        :param value: the stored object
        :type value: :class:`StringIO.StringIO` object
        :param timeout: the cache timeout in seconds
        """

    def get_last_modification(self, key):
        """Get last modification of cached file.

        :param key: the file object's key
        """

    def set_last_modification(self, key, last_modification=None, timeout=None):
        """Set last modification of cached file.

        :param key: the file object's key
        :param last_modification: Last modification date of
            file represented by the key
        :type last_modification: datetime.datetime
        :param timeout: the cache timeout in seconds
        """

    def _last_modification_key_name(self, key):
        """Generate key for last_modification entry of specified key.

        :param key: the file object's key
        """
        return "last_modification::%s" % key

    def delete(self, key):
        """Delete the specific key."""

    def flush(self):
        """Flush the cache."""

    def __call__(self, app=None):
        """Backwards-compatibility method returning ``self``."""
        return self



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2015, 2016 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Flask-IIIF decorators."""

from functools import wraps

from flask import current_app
from flask_restful import abort
from werkzeug.local import LocalProxy

from .errors import (
    IIIFValidatorError,
    MultimediaError,
    MultimediaImageCropError,
    MultimediaImageFormatError,
    MultimediaImageNotFound,
    MultimediaImageQualityError,
    MultimediaImageResizeError,
    MultimediaImageRotateError,
)

__all__ = (
    "api_decorator",
    "error_handler",
)

current_iiif = LocalProxy(lambda: current_app.extensions["iiif"])


def error_handler(f):
    """Error handler."""

    @wraps(f)
    def inner(*args, **kwargs):
        """Wrap the errors."""
        try:
            return f(*args, **kwargs)
        except (
            MultimediaImageCropError,
            MultimediaImageResizeError,
            MultimediaImageFormatError,
            MultimediaImageRotateError,
            MultimediaImageQualityError,
        ) as error:
            abort(500, message=error.message, code=500)
        except IIIFValidatorError as error:
            abort(400, message=error.message, code=400)
        except (MultimediaError, MultimediaImageNotFound) as error:
            abort(error.code, message=error.message, code=error.code)

    return inner


def api_decorator(f):
    """Decorate API method."""

    @wraps(f)
    def inner(*args, **kwargs):
        if current_iiif.api_decorator_callback:
            current_iiif.api_decorator_callback(*args, **kwargs)
        return f(*args, **kwargs)

    return inner



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2015, 2017 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Test REST API signals."""
import json

from .helpers import IIIFTestCase, signal_listener


class TestRestAPISignals(IIIFTestCase):

    """Test REST API signals."""

    def test_api_signals(self):
        """Test API signals."""
        from flask_iiif.signals import (
            iiif_after_info_request,
            iiif_after_process_request,
            iiif_before_info_request,
            iiif_before_process_request,
        )

        data = dict(
            uuid="valid:id",
            version="v2",
            region="full",
            size="full",
            rotation="0",
            quality="default",
            image_format="png",
        )

        infodata = dict(version="v2", uuid="valid:id")

        with signal_listener(iiif_before_process_request) as listener:
            self.get("iiifimageapi", urlargs=data)
            results = listener.assert_signal()
            self.assertEqual(results[1], data)

        with signal_listener(iiif_after_process_request) as listener:
            response = self.get("iiifimageapi", urlargs=data)
            results = listener.assert_signal()
            self.assertEqual(
                (results[1].get("mimetype"), results[1].get("image").getvalue()),
                ("image/png", response.data),
            )

        with signal_listener(iiif_before_info_request) as listener:
            response = self.get("iiifimageinfo", urlargs=infodata)
            results = listener.assert_signal()
            self.assertEqual(results[1], infodata)

        with signal_listener(iiif_after_info_request) as listener:
            response = self.get("iiifimageinfo", urlargs=infodata)
            results = listener.assert_signal()
            self.assertEqual(results[1], json.loads(response.get_data(as_text=True)))



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Image Simple Cache Tests."""

from __future__ import absolute_import

from io import BytesIO

from .helpers import IIIFTestCase


class TestImageSimpleCache(IIIFTestCase):

    """Multimedia Image Simple Cache test case."""

    def setUp(self):
        """Run before the test."""
        # Create a simple cache object
        from PIL import Image

        from flask_iiif.cache.simple import ImageSimpleCache

        # Initialize the cache object
        self.cache = ImageSimpleCache()
        # Create an image in memory
        tmp_file = BytesIO()
        # create a new image
        image = Image.new("RGBA", (1280, 1024), (255, 0, 0, 0))
        image.save(tmp_file, "png")
        # Store the image
        tmp_file.seek(0)
        self.image_file = tmp_file

    def test_set_and_get_function(self):
        """Test cache set and get functions."""
        # Seek position
        self.image_file.seek(0)
        # Add image to cache
        self.cache.set("image_1", self.image_file.getvalue())
        # Get image from cache
        image_string = self.cache.get("image_1")
        # test if the cache image is equal to the real
        self.assertEqual(image_string, self.image_file.getvalue())

    def test_image_recreation(self):
        """Test the image recreation from cache."""
        from flask_iiif.api import MultimediaImage

        # Seek position
        self.image_file.seek(0)
        # Add the image to cache
        self.cache.set("image_2", self.image_file.getvalue())
        # Get image from cache
        image_string = self.cache.get("image_2")
        # Create a ByteIO object
        cached_image = BytesIO(image_string)
        # Seek object to the right position
        cached_image.seek(0)
        # Create an image object form the stored string
        image = MultimediaImage.from_string(cached_image)
        # Check if the image is still the same
        self.assertEqual(str(image.size()), str((1280, 1024)))

    def test_cache_deletion(self):
        """Test cache delete function."""
        self.cache.set("foo", "bar")
        self.assertEqual(self.cache.get("foo"), "bar")
        self.cache.delete("foo")
        self.assertEqual(self.cache.get("foo"), None)

    def test_cache_flush(self):
        """Test cache flush function."""
        self.cache.set("foo_1", "bar")
        self.cache.set("foo_2", "bar")
        self.cache.set("foo_3", "bar")
        for i in [1, 2, 3]:
            self.assertEqual(self.cache.get("foo_{0}".format(i)), "bar")
        self.cache.flush()
        for i in [1, 2, 3]:
            self.assertEqual(self.cache.get("foo_{0}".format(i)), None)



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016, 2017 CERN.
# Copyright (C) 2020 data-futures.
# Copyright (C) 2022 Graz University of Technology.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Flask-IIIF extension provides easy IIIF API standard integration."""

from setuptools import setup

setup()



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015, 2016, 2017 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Implement a simple cache."""

from __future__ import absolute_import

from datetime import datetime

from cachelib.simple import SimpleCache

from .cache import ImageCache


class ImageSimpleCache(ImageCache):
    """Simple image cache."""

    def __init__(self, app=None):
        """Initialize the cache."""
        super(ImageSimpleCache, self).__init__(app=app)
        self.cache = SimpleCache()

    def get(self, key):
        """Return the key value.

        :param key: the object's key
        :return: the stored object
        :rtype: `BytesIO` object
        """
        return self.cache.get(key)

    def set(self, key, value, timeout=None):
        """Cache the object.

        :param key: the object's key
        :param value: the stored object
        :type value: `BytesIO` object
        :param timeout: the cache timeout in seconds
        """
        timeout = timeout or self.timeout
        self.cache.set(key, value, timeout)
        self.set_last_modification(key, timeout=timeout)

    def get_last_modification(self, key):
        """Get last modification of cached file.

        :param key: the file object's key
        """
        return self.get(self._last_modification_key_name(key))

    def set_last_modification(self, key, last_modification=None, timeout=None):
        """Set last modification of cached file.

        :param key: the file object's key
        :param last_modification: Last modification date of
            file represented by the key
        :type last_modification: datetime.datetime
        :param timeout: the cache timeout in seconds
        """
        if not last_modification:
            last_modification = datetime.utcnow().replace(microsecond=0)
        timeout = timeout or self.timeout
        self.cache.set(
            self._last_modification_key_name(key), last_modification, timeout
        )

    def delete(self, key):
        """Delete the specific key."""
        self.cache.delete(key)
        self.cache.delete(self._last_modification_key_name(key))

    def flush(self):
        """Flush the cache."""
        self.cache.clear()



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Multimedia error."""


class MultimediaError(Exception):
    """General multimedia exception."""

    def __init__(self, message=None, code=None):
        """Init the error handler."""
        super(MultimediaError, self).__init__()
        self.message = message or self.__class__.__name__
        self.code = code or 500

    def __str__(self):
        """Error message."""
        return repr(
            "Error message: {message}. Error code: {code}".format(
                message=self.message, code=self.code
            )
        )


class MultimediaImageNotFound(MultimediaError):
    """Image not found error."""

    def __init__(self, message=None):
        """Init with status code 404."""
        super(MultimediaImageNotFound, self).__init__(message, code=404)


class MultimediaImageCropError(MultimediaError):
    """Image on crop error."""


class MultimediaImageResizeError(MultimediaError):
    """Image resize error."""


class MultimediaImageRotateError(MultimediaError):
    """Image rotate error."""


class MultimediaImageQualityError(MultimediaError):
    """Image quality error."""


class MultimediaImageFormatError(MultimediaError):
    """Image format error."""


class IIIFValidatorError(MultimediaError):
    """IIIF API validator error."""



# -*- coding: utf-8 -*-
#
# This file is part of Flask-IIIF
# Copyright (C) 2014, 2015 CERN.
#
# Flask-IIIF is free software; you can redistribute it and/or modify
# it under the terms of the Revised BSD License; see LICENSE file for
# more details.

"""Flask-IIIF extension test."""

from __future__ import absolute_import

from unittest import TestCase

from flask import Flask

from flask_iiif import IIIF
from flask_iiif import config as default_config


class TestIIIF(TestCase):

    """Test extension creation."""

    def setUp(self):
        """Setup up."""
        app = Flask(__name__)
        app.config["DEBUG"] = True
        app.config["TESTING"] = True
        app.logger.disabled = True
        self.app = app

    def test_version(self):
        """Assert that version number can be parsed."""
        from distutils.version import LooseVersion

        from flask_iiif import __version__

        LooseVersion(__version__)

    def test_creation(self):
        """Test extension creation."""
        assert "iiif" not in self.app.extensions
        IIIF(app=self.app)
        assert isinstance(self.app.extensions["iiif"], IIIF)

    def test_creation_old_flask(self):
        """Simulate old Flask (pre 0.9)."""
        del self.app.extensions
        IIIF(app=self.app)
        assert isinstance(self.app.extensions["iiif"], IIIF)

    def test_creation_init(self):
        """Test extension creation init."""
        assert "iiif" not in self.app.extensions
        r = IIIF()
        r.init_app(app=self.app)
        assert isinstance(self.app.extensions["iiif"], IIIF)

    def test_double_creation(self):
        """Test extension double creation."""
        IIIF(app=self.app)
        self.assertRaises(RuntimeError, IIIF, app=self.app)

    def test_default_config(self):
        """Test extension default configuration."""
        IIIF(app=self.app)
        for k in dir(default_config):
            if k.startswith("IIIF_"):
                assert self.app.config.get(k) == getattr(default_config, k)
